import json
import logging
import subprocess

from kamikaze3.autoscaling.varnish.main import (
    run_validations as run_varnish_validations,
)
from kamikaze3.common.args import OUTPUT_FORMAT_JSON
from kamikaze3.common.autoscaling import (
    HORIZONTAL_AUTOSCALING_REQUIREMENTS_DOCUMENTATION_LINK,
    HYPERNODE_SPECIFIC_REQUIREMENTS_DOCUMENTATION_LINK,
    MYSQL_VERSION_REQUIREMENT_DOCUMENTATION_LINK,
    OS_REQUIREMENT_DOCUMENTATION_LINK,
    PODMAN_REQUIREMENT_DOCUMENTATION_LINK,
    PRODUCTION_REQUIREMENT_DOCUMENTATION_LINK,
    REDIS_PERSISTENT_REQUIREMENT_DOCUMENTATION_LINK,
    SUPERVISOR_REQUIREMENT_DOCUMENTATION_LINK,
    VARNISH_REQUIREMENT_DOCUMENTATION_LINK,
)
from kamikaze3.common.env import get_hn_app_var, get_hn_common_var
from kamikaze3.common.system import get_os_release
from kamikaze3.event_listener.settings import PROGRAM_NAME
from kamikaze3.supervisorctl.service import get_process_statuses, is_supervisor_running

logger = logging.getLogger(PROGRAM_NAME)

HYPERNODE_ELIGIBILITY_CHECKS = []

DISALLOWED_MYSQL_VERSIONS = ["5.6"]
DISALLOWED_OPERATING_SYSTEMS = ["buster"]
SUPPORTED_PROVIDERS = ["FALCON"]


def hypernode_check(
    name,
    description,
    documentation_link=HYPERNODE_SPECIFIC_REQUIREMENTS_DOCUMENTATION_LINK,
):
    def wrapper(func):
        HYPERNODE_ELIGIBILITY_CHECKS.append(
            {
                "name": name,
                "description": description,
                "function": func,
                "documentation_link": documentation_link,
            }
        )
        return func

    return wrapper


@hypernode_check(
    name="is_not_cluster",
    description="Checking if the Hypernode is not part of a cluster",
)
def is_not_cluster():
    cluster_var = get_hn_common_var("cluster")
    if cluster_var is not None and cluster_var.get("id"):
        if cluster_var.get("type") == "autoscale":
            return True, "Hypernode is in an autoscale cluster, which is allowed"
        else:
            return False, "Hypernode is a part of a cluster"

    return True, "Hypernode is not part of a cluster"


@hypernode_check(
    name="is_varnish_enabled",
    description="Checking if Varnish is enabled",
    documentation_link=VARNISH_REQUIREMENT_DOCUMENTATION_LINK,
)
def is_varnish_enabled():
    common_var = get_hn_common_var("varnish")
    if common_var is not None and common_var.get("enabled"):
        return True, "Varnish is enabled"

    return False, "Varnish is not enabled"


@hypernode_check(
    name="is_production",
    description="Checking if its a production Hypernode",
    documentation_link=PRODUCTION_REQUIREMENT_DOCUMENTATION_LINK,
)
def is_production():
    if not get_hn_app_var("is_development"):
        return True, "Hypernode is a production environment"

    return False, "Hypernode is not a production environment"


@hypernode_check(
    name="is_redis_persistent_instance",
    description="Checking if the Redis instance is persistent",
    documentation_link=REDIS_PERSISTENT_REQUIREMENT_DOCUMENTATION_LINK,
)
def is_redis_persistent_instance():
    if get_hn_app_var("redis_persistent_instance"):
        return True, "Redis instance is persistent"

    return False, "Redis instance is not persistent"


@hypernode_check(
    name="validate_mysql_version",
    description="Checking if the MySQL version is allowed",
    documentation_link=MYSQL_VERSION_REQUIREMENT_DOCUMENTATION_LINK,
)
def validate_mysql_version():
    if not get_hn_app_var("mysql_version") in DISALLOWED_MYSQL_VERSIONS:
        return True, "MySQL version is allowed."

    return False, "MySQL version is not allowed."


@hypernode_check(
    name="is_provider_supported",
    description="Checking if the Hypernode provider is supported",
    documentation_link=HORIZONTAL_AUTOSCALING_REQUIREMENTS_DOCUMENTATION_LINK,
)
def is_provider_supported():
    product = get_hn_app_var("product")
    if product and any(
        product.startswith(provider) for provider in SUPPORTED_PROVIDERS
    ):
        return True, "Provider is supported."

    return False, "Only Falcon Plans are supported for now."


@hypernode_check(
    name="validate_operating_system",
    description="Checking if the operating system is allowed",
    documentation_link=OS_REQUIREMENT_DOCUMENTATION_LINK,
)
def validate_operating_system():
    if not get_os_release() in DISALLOWED_OPERATING_SYSTEMS:
        return True, "Operating system is allowed."

    return False, "Operating system is not allowed."


@hypernode_check(
    name="validate_varnish_configuration",
    description="Checking if the Varnish configuration is valid",
    documentation_link=VARNISH_REQUIREMENT_DOCUMENTATION_LINK,
)
def validate_varnish_configuration():
    result = False, "Varnish configuration is not valid"
    try:
        varnish_enabled, _ = is_varnish_enabled()

        if varnish_enabled:
            response = run_varnish_validations(OUTPUT_FORMAT_JSON)
            result = json.loads(response)
        else:
            result = False, "Varnish is not enabled and configured"
    except OSError as e:
        logger.error(
            f"OSError: Could not validate Varnish status and configuration: {e}"
        )

    return result


@hypernode_check(
    name="validate_supervisor_configuration",
    description="Checking supervisor is disabled and that no supervisor services are configured",
    documentation_link=SUPERVISOR_REQUIREMENT_DOCUMENTATION_LINK,
)
def validate_supervisor_configuration():
    supervisor_running = is_supervisor_running()

    if supervisor_running:
        active_procs = list(
            filter(lambda proc: proc.is_active(), get_process_statuses())
        )
        if active_procs:
            for process in active_procs:
                return (
                    False,
                    f"Supervisor service {process.name} is {process.state.lower()}. Please disable all the processes.",
                )

    return True, "There are no services configured in the supervisor"


@hypernode_check(
    name="validate_podman_configuration",
    description="Checking podman is disabled or no podman services are running",
    documentation_link=PODMAN_REQUIREMENT_DOCUMENTATION_LINK,
)
def validate_podman_configuration():
    podman_ps = subprocess.check_output(
        "/usr/bin/sudo -u app /usr/bin/podman ps --format json",
        shell=True,
        cwd="/data/web",
        encoding="utf-8",
    )

    containers = json.loads(podman_ps)
    if containers:
        container_names = []
        for c in containers:
            for name in c.get("Names", []):
                container_names.append(name)
                break
        return (
            False,
            f"Podman is running the following containers: {container_names}",
        )

    return True, "There are no services configured in Podman"
