import json
import logging
import os
import subprocess
from pathlib import Path

from kamikaze3.analysis.app_locator import (
    APP_VERSION_UNKNOWN,
    ApplicationInfo,
    detect_composer_app,
)
from kamikaze3.autoscaling.eligibility.application_checks import application_check
from kamikaze3.common.autoscaling import (
    CMS_REQUIREMENT_DOCUMENTATION_LINK,
    HOSTNAMES_REQUIREMENT_DOCUMENTATION_LINK,
    OPENSEARCH_REQUIREMENT_DOCUMENTATION_LINK,
    RABBITMQ_REQUIREMENT_DOCUMENTATION_LINK,
    REDIS_PERSISTENT_REQUIREMENT_DOCUMENTATION_LINK,
    REMOTE_STORAGE_REQUIREMENT_DOCUMENTATION_LINK,
    VARNISH_REQUIREMENT_DOCUMENTATION_LINK,
    REDIS_CACHE_REQUIREMENT_DOCUMENTATION_LINK,
)
from kamikaze3.common.version import LooseVersion

logger = logging.getLogger(__name__)

SEARCH_ENGINE_OPENSEARCH = "opensearch"
SEARCH_ENGINE_ELASTICSEARCH7 = "elasticsearch7"
MIN_MAGENTO_VERSION = "2.4.7"

ALLOWED_REDIS_HOST_NAMES = ["redismaster"]
REDIS_HOST_PORT = "6379"


class Magento2Information(ApplicationInfo):
    def __init__(
        self,
        path: os.PathLike,
        enabled_modules: list[str] | None = None,
        env_config: dict | None = None,
        global_config: dict | None = None,
        version: str = APP_VERSION_UNKNOWN,
    ):
        super().__init__(path, "magento2", version=version)
        self.enabled_modules = enabled_modules or []
        self.env_config = env_config or {}
        self.global_config = global_config or {}


@application_check(
    name="magento2_remote_storage_enabled",
    application="magento2",
    description="Check if Magento_RemoteStorage module is enabled",
    documentation_link=REMOTE_STORAGE_REQUIREMENT_DOCUMENTATION_LINK,
)
def remote_storage_enabled(app_info: Magento2Information) -> tuple[bool, str]:
    """
    Validate if Magento_RemoteStorage module is enabled
    """
    if "Magento_RemoteStorage" in app_info.enabled_modules:
        return True, "Magento_RemoteStorage is enabled"
    return False, "Magento_RemoteStorage is not enabled"


@application_check(
    name="magento2_database_storage_disabled",
    application="magento2",
    description="Check if database storage is disabled",
    documentation_link=REMOTE_STORAGE_REQUIREMENT_DOCUMENTATION_LINK,
)
def database_storage_disabled(app_info: Magento2Information) -> tuple[bool, str]:
    """
    Validate if remote database storage is disabled
    """
    media_storage_enabled = int(
        app_info.global_config.get(
            "system/media_storage_configuration/media_database", "0"
        )
    )
    if media_storage_enabled:
        return (
            False,
            "system/media_storage_configuration/media_database is not disabled",
        )
    return True, "Database storage is disabled"


@application_check(
    name="magento2_remote_storage_configured",
    application="magento2",
    description="Check if remote storage is configured correctly",
    documentation_link=REMOTE_STORAGE_REQUIREMENT_DOCUMENTATION_LINK,
)
def remote_storage_configured_correctly(
    app_info: Magento2Information,
) -> tuple[bool, str]:
    """
    Validate if app/etc/env.php configured for remote storage
    """
    remote_storage = app_info.env_config.get("remote_storage")
    if not remote_storage:
        return (
            False,
            "'remote_storage' key is missing or misconfigured in app/etc/env.php",
        )

    driver = remote_storage.get("driver", "").lower()
    if driver not in ["aws-s3"]:
        return (
            False,
            "Make sure your application uses object storage with a supported remote storage driver",
        )

    config = remote_storage.get("config", {})
    required_fields = ["bucket", "region", "credentials"]

    missing_fields = [field for field in required_fields if not config.get(field)]
    if missing_fields:
        return (
            False,
            f"Missing required fields: {', '.join(missing_fields)} in remote storage configuration",
        )

    return True, "Remote Storage is configured correctly"


@application_check(
    name="magento2_hostnames_configured_correctly",
    application="magento2",
    description="Check if hostnames are configured correctly",
    documentation_link=HOSTNAMES_REQUIREMENT_DOCUMENTATION_LINK,
)
def hostnames_configured_correctly(app_info: Magento2Information) -> tuple[bool, str]:
    """
    Validate if app/etc/env.php configured for hostnames, not IPs
    """
    # Potential keys in env.php
    # (https://experienceleague.adobe.com/en/docs/commerce-operations/configuration-guide/files/config-reference-envphp)
    # backend, cache, cache_types, consumers_wait_for_messages, cron, crypt, db, default_connection, directories,
    # downloadable_domains, install, lock, MAGE_MODE, queue, resource, session, system, x-frame-options
    scan_keys = ["db", "cache", "session", "queue"]
    for scan in scan_keys:
        section = app_info.env_config.get(scan)
        if not section:
            continue

        result, message = scan_connections(section, scan)
        if not result:
            return result, message

    return True, "Hostnames are configured correctly"


@application_check(
    name="magento2_rabbitmq_configured_without_guest_user",
    application="magento2",
    description="Check if RabbitMQ is configured without guest user for external connections",
    documentation_link=RABBITMQ_REQUIREMENT_DOCUMENTATION_LINK,
)
def rabbitmq_configured_without_guest_user(
    app_info: Magento2Information,
) -> tuple[bool, str]:
    """
    Validate if app/etc/env.php configured for RabbitMQ without guest user for external connections
    """
    queue = app_info.env_config.get("queue")

    if not queue:
        return (
            True,
            "RabbitMQ is not being used",
        )

    config = queue.get("amqp", {})
    if config.get("user") == "guest" and config.get("password") == "guest":
        return (
            False,
            f"Please configure RabbitMQ user which is enabled for external connections",
        )

    return True, "RabbitMQ user credentials is configured correctly"


def scan_connections(config, parent_key=""):
    if isinstance(config, dict):
        potential_keys = ["host", "server"]
        invalid_hosts = ["localhost", "127.0.0.1"]
        for key, value in config.items():
            current_path = f"{parent_key}.{key}" if parent_key != "" else key
            if key in potential_keys:
                host_value_invalid = value in invalid_hosts or "/" in value
                if host_value_invalid:
                    return (
                        False,
                        f"Invalid host '{value}' found at '{current_path}'",
                    )
            _result, _message = scan_connections(value, current_path)
            if not _result:
                return _result, _message
    elif isinstance(config, list):
        for index, item in enumerate(config):
            current_path = f"{parent_key}[{index}]"
            _result, _message = scan_connections(item, current_path)
            if not _result:
                return _result, _message
    return True, None


@application_check(
    name="magento2_opensearch_configured_correctly",
    application="magento2",
    description="Check if OpenSearch hostname and port are configured correctly",
    documentation_link=OPENSEARCH_REQUIREMENT_DOCUMENTATION_LINK,
)
def opensearch_configured_correctly(app_info: Magento2Information) -> tuple[bool, str]:
    expected_hostname = "elasticsearchmaster"
    expected_port = 9200

    search_engine = app_info.global_config.get("catalog/search/engine")
    if not search_engine and app_info.version != APP_VERSION_UNKNOWN:
        # OpenSearch became the default search engine in 2.4.6-p1:
        # https://github.com/magento/magento2/commit/ebe8018dec4b7a8a61e4741d75a839a1098b1f84
        if LooseVersion(app_info.version) >= LooseVersion("2.4.6-p1"):
            search_engine = SEARCH_ENGINE_OPENSEARCH
        else:
            search_engine = SEARCH_ENGINE_ELASTICSEARCH7

    if search_engine == SEARCH_ENGINE_OPENSEARCH:
        hostname = app_info.global_config.get(
            "catalog/search/opensearch_server_hostname"
        )
        port = int(
            app_info.global_config.get("catalog/search/opensearch_server_port", "9200")
        )
        if hostname != expected_hostname:
            return False, "OpenSearch hostname configured incorrectly"
        if port != expected_port:
            return False, "OpenSearch port configured incorrectly"
        return True, "OpenSearch configured correctly"
    elif search_engine == SEARCH_ENGINE_ELASTICSEARCH7:
        hostname = app_info.global_config.get(
            "catalog/search/elasticsearch7_server_hostname"
        )
        port = int(
            app_info.global_config.get(
                "catalog/search/elasticsearch7_server_port", "9200"
            )
        )
        if hostname != expected_hostname:
            return False, "ElasticSearch hostname configured incorrectly"
        if port != expected_port:
            return False, "ElasticSearch port configured incorrectly"
        return True, "ElasticSearch configured correctly"
    else:
        return False, "Unknown search engine configured"


@application_check(
    name="magento2_varnish_http_cache_hosts_configured_correctly",
    application="magento2",
    description="Check if Varnish HTTP cache host is configured correctly",
    documentation_link=VARNISH_REQUIREMENT_DOCUMENTATION_LINK,
)
def varnish_http_cache_hosts_configured_correctly(
    app_info: Magento2Information,
) -> tuple[bool, str]:
    """
    Validate if app/etc/env.php is configured for Varnish http_cache_hosts with external connections.
    This function checks for any host entries other than 'localhost' or '127.0.0.1'.
    """
    http_cache_hosts = app_info.env_config.get("http_cache_hosts")

    if not http_cache_hosts:
        return True, "http_cache_hosts is not set."

    for host_info in http_cache_hosts:
        host = host_info.get("host", "")
        if host not in ("localhost", "127.0.0.1"):
            return True, f"External host found: {host}"

    return (
        False,
        "No external http_cache_host configured in app config; only localhost or 127.0.0.1 found.",
    )


@application_check(
    name="magento2_redis_sessions_configured_correctly",
    application="magento2",
    description="Check if Redis session is configured correctly",
    documentation_link=REDIS_PERSISTENT_REQUIREMENT_DOCUMENTATION_LINK,
)
def redis_sessions_configured_correctly(
    app_info: Magento2Information,
) -> tuple[bool, str]:
    """
    Validate if app/etc/env.php configured for Redis sessions
    """
    session = app_info.env_config.get("session")
    if not session:
        return False, "Redis Session configuration is missing"

    save = session.get("save")
    if save != "redis":
        return False, "Session save is not set to redis"

    redis = session.get("redis")
    if not redis:
        return False, "Redis configuration is missing"

    host = redis.get("host")
    if host not in ALLOWED_REDIS_HOST_NAMES:
        return False, "Redis host is not set to redismaster"

    port = redis.get("port")
    if port != "6378":
        return False, "Redis port is not set to 6378"

    return True, "Redis session is configured correctly"


@application_check(
    name="magento2_redis_cache_frontend_configured_correctly",
    application="magento2",
    description="Check if Redis cache frontend is configured correctly",
    documentation_link=REDIS_CACHE_REQUIREMENT_DOCUMENTATION_LINK,
)
def redis_cache_frontend_configured_correctly(
    app_info: Magento2Information,
) -> tuple[bool, str]:
    """
    Validate if app/etc/env.php configured for Redis sessions
    """
    cache = app_info.env_config.get("cache")
    if not cache:
        return False, "Redis cache configuration is missing"

    frontend = cache.get("frontend")
    if not frontend:
        return False, "Redis cache frontend configuration is missing"

    for cache_type in frontend.keys():
        cache_frontend = frontend.get(cache_type)
        backend_options = cache_frontend.get("backend_options")
        if not backend_options:
            return (
                False,
                f"backend_options section missing for redis cache type {cache_type}",
            )

        host = backend_options.get("server")
        if not host or host not in ALLOWED_REDIS_HOST_NAMES:
            return (
                False,
                f"Redis cache host is not set to redismaster for cache type {cache_type}",
            )

        port = backend_options.get("port")
        if not port or port != REDIS_HOST_PORT:
            return (
                False,
                f"Redis cache host port is not set to {REDIS_HOST_PORT} for cache type {cache_type}",
            )

    return True, "Redis frontend cache is configured correctly"


@application_check(
    name="magento2_version_check",
    application="magento2",
    description="Check Magento version is 2.4.7 or higher",
    documentation_link=CMS_REQUIREMENT_DOCUMENTATION_LINK,
)
def version_check(
    app_info: Magento2Information,
) -> tuple[bool, str]:
    """
    Validate Magento2 version is 2.4.7 or higher
    """
    if app_info.version == APP_VERSION_UNKNOWN:
        return False, f"Could not determine Magento 2 version"

    if LooseVersion(app_info.version) < LooseVersion(MIN_MAGENTO_VERSION):
        return False, f"Magento2 version is lower than {MIN_MAGENTO_VERSION}"

    return True, f"Your Magento2 version is {MIN_MAGENTO_VERSION} or higher"


def collect_magento2_information(path: Path) -> Magento2Information:
    """
    Collect Magento2 information
    """
    try:
        module_status_output = subprocess.check_output(
            f"/usr/bin/sudo -u app /usr/bin/php bin/magento module:status --enabled",
            shell=True,
            cwd=path,
            encoding="utf-8",
        )
        enabled_modules = [m.strip() for m in module_status_output.splitlines()]
    except subprocess.CalledProcessError as exc:
        logger.error("Could not retrieve enabled modules: %s", exc)
        enabled_modules = []

    try:
        env_config_output = subprocess.check_output(
            f"/usr/bin/sudo -u app /usr/bin/php -r \"echo json_encode(include 'app/etc/env.php');\"",
            shell=True,
            cwd=path,
            encoding="utf-8",
        )
        env_config = json.loads(env_config_output) if env_config_output.strip() else {}
    except subprocess.CalledProcessError as exc:
        logger.error("Could not retrieve env.php config: %s", exc)
        env_config = {}

    global_config = {}
    try:
        global_config_output = subprocess.check_output(
            f"/usr/bin/sudo -u app /usr/bin/php bin/magento config:show --scope default --scope-code ''",
            shell=True,
            cwd=path,
            encoding="utf-8",
        )
        last_index = None
        for line in global_config_output.splitlines():
            try:
                key, value = line.split("-", maxsplit=1)
                last_index = key.strip()
                global_config[last_index] = value.strip()
            except ValueError:
                if last_index:
                    global_config[last_index] += "\n" + line.strip()
                else:
                    logger.warning(
                        "Could not parse global config output line: %s", line
                    )
                    pass
    except subprocess.CalledProcessError as exc:
        logger.error("Could not retrieve global config: %s", exc)
        global_config = {}

    composer_app = detect_composer_app(Path(path))

    return Magento2Information(
        path,
        enabled_modules,
        env_config,
        global_config,
        version=composer_app.version if composer_app else APP_VERSION_UNKNOWN,
    )
