import asyncio
import json
import logging
from collections.abc import Awaitable, Callable
from pathlib import Path
from subprocess import CalledProcessError, check_output

from nats.aio.msg import Msg

from kamikaze3 import ftp
from kamikaze3.analysis import app_locator
from kamikaze3.analysis.app_locator import ApplicationInfo
from kamikaze3.autoscaling import APPLICATION_INFORMATION_COLLECTORS
from kamikaze3.autoscaling.eligibility.application_checks import (
    APPLICATION_ELIGIBILITY_CHECKS,
)
from kamikaze3.autoscaling.eligibility.hypernode_checks import (
    HYPERNODE_ELIGIBILITY_CHECKS,
)
from kamikaze3.common.env import get_hn_common_var
from kamikaze3.dbus.interfaces.nginx_config_reloader import NGINX_CONFIG_RELOADER
from kamikaze3.event_listener.client import create_response
from kamikaze3.event_listener.settings import PROGRAM_NAME

logger = logging.getLogger(PROGRAM_NAME)


def event_handler_wrapper(
    handler: Callable[[Msg], Awaitable],
) -> Callable[[Msg], Awaitable]:
    async def _wrapper(msg: Msg):
        logger.info(f"Received event {msg.subject}")
        resp = await handler(msg)
        logger.debug(f"Handled event {msg.subject}")
        return resp

    return _wrapper


async def reload_hmv(_: Msg):
    logger.debug("Reloading HMV")
    try:
        check_output(
            [
                "/usr/bin/sudo",
                "-uapp",
                "/usr/bin/hypernode-manage-vhosts",
                "--no",
                "--all",
                "--ssl-noclobber",
                "--skip-event",
            ]
        )
    except CalledProcessError as e:
        logger.error(f"Failed to reload HMV: {str(e)}")
        return
    logger.debug("Done reloading HMV")


async def list_vhosts(msg: Msg):
    vhosts = check_output(
        r"/usr/sbin/nginx -T 2>&1 | /usr/bin/grep -E '^\s*server_name ' | /usr/bin/awk '{print$2}' | /usr/bin/cut -d';' -f1 | /usr/bin/sort | /usr/bin/uniq",
        shell=True,
    )
    logger.debug(f"Found vhosts: {vhosts.decode()}")
    return await msg.respond(vhosts)


async def show_fpm_status(msg: Msg):
    status = check_output("/usr/bin/hypernode-fpm-status", shell=True)
    logger.debug(f"Found FPM status: {status.decode()}")
    return await msg.respond(status)


async def show_ftp_users(msg: Msg):
    try:
        users = ftp.list_users()
    except OSError:
        logger.error("IOError: Could not list FTP users")
        return await msg.respond(
            create_response(message="Could not list FTP users", status=1)
        )

    logger.debug(f"Found FTP users: {users}")
    return await msg.respond(create_response(message=None, status=0, data=users))


async def add_ftp_user(msg: Msg):
    try:
        data = json.loads(msg.data.decode())
    except json.JSONDecodeError as e:
        logger.error(f"FTP user could not be added due {str(e)}")
        return await msg.respond(
            create_response(message="Ftp user could not be added", status=1)
        )

    for required_field in ["user", "password", "homedir"]:
        if required_field in data:
            continue
        logger.error(f"FTP user could not be added. Missing {required_field}")
        return await msg.respond(
            create_response(
                message=f"FTP user could not be added. Missing {required_field}",
                status=1,
            )
        )

    try:
        user = ftp.add_user(data["user"], data["password"], data["homedir"])
    except SystemExit as e:
        logger.error(f"FTP user could not be added due {str(e)}")
        return await msg.respond(create_response(message=str(e), status=1))

    logger.info(f"FTP user added: {data['user']} for {data['homedir']}")
    return await msg.respond(create_response(message=None, status=0, data=user))


async def remove_ftp_user(msg: Msg):
    try:
        data = json.loads(msg.data.decode())
    except json.JSONDecodeError as e:
        logger.error(f"FTP user could not be removed due {str(e)}")
        return await msg.respond(
            create_response(message="Ftp user could not be removed", status=1)
        )

    if "user" not in data:
        logger.error("FTP user could not be removed. Missing username in msg")
        return await msg.respond(
            create_response(
                message="FTP user could not be removed, field user is required",
                status=1,
            )
        )

    try:
        user = ftp.remove_user(data["user"])
    except SystemExit as e:
        logger.error(f"FTP user could not be removed due {str(e)}")
        return await msg.respond(create_response(message=str(e), status=1))

    logger.info(f"FTP user removed: {data['user']}")
    return await msg.respond(create_response(message=None, status=0, data=user))


# Useful for debugging
async def pong(msg: Msg):
    logger.debug(f"Received ping from {msg.subject}")
    return await msg.respond(b"pong")


async def reload_nginx(msg: Msg):
    logger.debug("Triggering nginx reload")
    nginx_config_reloader_proxy = NGINX_CONFIG_RELOADER.get_proxy()
    nginx_config_reloader_proxy.Reload()
    return await msg.respond(b"reloaded")


async def find_applications(msg: Msg):
    try:
        result = await asyncio.to_thread(
            app_locator.find_applications, Path("/data/web")
        )
        installed_apps = json.dumps(list(map(lambda x: x.serialize(), result)))
    except Exception as e:
        logger.error(f"Failed to locate installed applications: {str(e)}")
        return await msg.respond(
            create_response(message="Failed to locate installed applications", status=1)
        )

    logger.debug(f"Found installed applications: {len(result)}")
    return await msg.respond(
        create_response(message=None, status=0, data=installed_apps)
    )


async def check_application_eligibility_for_horizontal_autoscaling(msg: Msg):
    try:
        data = json.loads(msg.data.decode())
    except json.JSONDecodeError as e:
        logger.error(f"Application eligibility check for autoscaling failed: {str(e)}")
        return await msg.respond(
            create_response(
                message="Application eligibility for autoscaling couldnt be checked",
                status=1,
            )
        )

    app_type = data.get(
        "application_type", data.get("application")
    )  # backwards compatibility
    if not app_type or app_type not in APPLICATION_ELIGIBILITY_CHECKS:
        return await msg.respond(
            create_response(
                message="Application eligibility for autoscaling couldn't be checked, invalid application",
                status=1,
            )
        )

    collector = APPLICATION_INFORMATION_COLLECTORS[app_type]
    app_info = await asyncio.to_thread(collector, data["path"])

    result = []
    for check in APPLICATION_ELIGIBILITY_CHECKS[app_type]:
        check_fn: Callable[[ApplicationInfo], tuple[bool, str]] = check["function"]
        try:
            check_response = check_fn(app_info)
        except Exception as e:
            logger.error(
                f"Got exception for application eligibility check {check['name']}: {str(e)}"
            )
            check_response = (
                False,
                "Something went wrong executing this eligibility check",
            )
        result.append(
            {
                "name": check["name"],
                "description": check["description"],
                "documentation_link": check["documentation_link"],
                "checks": check_response,
            }
        )

    logger.debug(f"Application eligibility check status: {json.dumps(result)}")

    response = create_response(message=None, status=0, data=result)

    return await msg.respond(response)


async def check_hypernode_eligibility_for_horizontal_autoscaling(msg: Msg):
    result = []
    for check in HYPERNODE_ELIGIBILITY_CHECKS:
        try:
            check_response = check["function"]()
        except Exception as e:
            logger.error(
                f"Got exception for hypernode eligibility check {check['name']}: {str(e)}"
            )
            check_response = (
                False,
                "Something went wrong executing this eligibility check",
            )
        result.append(
            {
                "name": check["name"],
                "description": check["description"],
                "documentation_link": check["documentation_link"],
                "checks": check_response,
            }
        )

    logger.debug(f"Hypernode eligibility check status: {json.dumps(result)}")

    response = create_response(message=None, status=0, data=result)

    return await msg.respond(response)


async def show_autoscale_status(msg: Msg):
    status = []
    try:
        check_output(["/usr/bin/systemctl", "is-active", "--quiet", "haproxy"])
    except CalledProcessError:
        status.append("Haproxy service not available and ready for use.")

    cluster_ip = get_hn_common_var("cluster_ip", None)
    if not cluster_ip:
        status.append("Cluster IP not configured")
    else:
        try:
            check_output(["/usr/bin/ping", "-c1", cluster_ip])
        except CalledProcessError:
            status.append(f"Unable to ping configured cluster_ip {cluster_ip}")

    response_data = {
        "status": "1" if not status else "0",
        "data": status,
    }

    return await msg.respond(json.dumps(response_data).encode(errors="ignore"))


def remove_func_from_eligibility_check_dict(check_dict: dict):
    return {
        "name": check_dict["name"],
        "description": check_dict["description"],
        "documentation_link": check_dict["documentation_link"],
    }


async def autoscaling_eligibility_checks(msg: Msg):
    supported_application_types = []
    application_checks = []
    hypernode_checks = []

    for application_type, eligibility_checks in APPLICATION_ELIGIBILITY_CHECKS.items():
        application_checks.append(
            {
                application_type: [
                    remove_func_from_eligibility_check_dict(check)
                    for check in eligibility_checks
                ]
            }
        )
        supported_application_types.append(application_type)

    for hypernode_check in HYPERNODE_ELIGIBILITY_CHECKS:
        hypernode_checks.append(
            remove_func_from_eligibility_check_dict(hypernode_check)
        )

    eligibility_checks = {
        "hypernode_checks": hypernode_checks,
        "application_checks": application_checks,
        # @TODO(timon): this ambiguous key should be removed later:
        "autoscalable_applications": supported_application_types,
        "supported_application_types": supported_application_types,
    }

    response = create_response(message=None, status=0, data=eligibility_checks)

    return await msg.respond(response)
