import argparse
import asyncio
from collections.abc import Awaitable, Callable
from logging import getLogger
from signal import SIGINT, SIGTERM

from nats.aio.client import Client
from nats.aio.msg import Msg
from nats.aio.subscription import Subscription

from kamikaze3.common.env import get_hypernode_name
from kamikaze3.common.log import setup_logging
from kamikaze3.event_listener.client import get_nats_client
from kamikaze3.event_listener.events import (
    add_ftp_user,
    autoscaling_eligibility_checks,
    check_application_eligibility_for_horizontal_autoscaling,
    check_hypernode_eligibility_for_horizontal_autoscaling,
    event_handler_wrapper,
    find_applications,
    list_vhosts,
    pong,
    reload_hmv,
    reload_nginx,
    remove_ftp_user,
    show_autoscale_status,
    show_fpm_status,
    show_ftp_users,
)
from kamikaze3.event_listener.settings import (
    EVENT_ADD_FTP_USER,
    EVENT_AUTOSCALING_APPLICATION_ELIGIBLE,
    EVENT_AUTOSCALING_ELIGIBILITY_CHECKS,
    EVENT_AUTOSCALING_HYPERNODE_ELIGIBLE,
    EVENT_FIND_APPLICATIONS,
    EVENT_LIST_VHOSTS,
    EVENT_PING,
    EVENT_RELOAD_HMV,
    EVENT_RELOAD_NGINX,
    EVENT_REMOVE_FTP_USER,
    EVENT_SHOW_AUTOSCALE_STATUS,
    EVENT_SHOW_FPM_STATUS,
    EVENT_SHOW_FTP_USERS,
    PROGRAM_NAME,
)
from kamikaze3.event_listener.utils import get_default_nats_server

logger = getLogger(PROGRAM_NAME)

EVENT_MAPPING: dict[str, Callable[[Msg], Awaitable]] = {
    EVENT_RELOAD_HMV: reload_hmv,
    EVENT_RELOAD_NGINX: reload_nginx,
    EVENT_LIST_VHOSTS: list_vhosts,
    EVENT_SHOW_FPM_STATUS: show_fpm_status,
    EVENT_SHOW_FTP_USERS: show_ftp_users,
    EVENT_ADD_FTP_USER: add_ftp_user,
    EVENT_REMOVE_FTP_USER: remove_ftp_user,
    EVENT_PING: pong,
    EVENT_FIND_APPLICATIONS: find_applications,
    EVENT_SHOW_AUTOSCALE_STATUS: show_autoscale_status,
    EVENT_AUTOSCALING_ELIGIBILITY_CHECKS: autoscaling_eligibility_checks,
    EVENT_AUTOSCALING_HYPERNODE_ELIGIBLE: check_hypernode_eligibility_for_horizontal_autoscaling,
    EVENT_AUTOSCALING_APPLICATION_ELIGIBLE: check_application_eligibility_for_horizontal_autoscaling,
}


def parse_args(args: list[str] | None = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Start the Hypernode event listener")
    parser.add_argument(
        "-s", "--server", help="The NATS server to connect to", required=False
    )
    parser.add_argument(
        "-v", "--verbose", help="Enable verbose logging", action="store_true"
    )
    return parser.parse_args(args)


async def stop(nc: Client):
    logger.debug("Draining NATS connection...")
    await nc.drain()
    logger.info(f"Closing NATS connection")
    await nc.close()
    asyncio.get_running_loop().stop()


async def listener_loop(server: str):
    logger.info(f"Starting Hypernode event listener loop for {server}")
    nc = await get_nats_client(server, get_hypernode_name())

    subs: list[Subscription] = []

    def signal_handler():
        if nc.is_closed:
            return
        asyncio.create_task(stop(nc))

    for sig in (SIGINT, SIGTERM):
        asyncio.get_running_loop().add_signal_handler(sig, signal_handler)

    for event, handler in EVENT_MAPPING.items():
        logger.debug(f"Subscribing to event {event}")
        sub = await nc.subscribe(event, cb=event_handler_wrapper(handler=handler))
        subs.append(sub)


def get_server(args):
    if args.server:
        return args.server
    return get_default_nats_server()


def main(initial_args: list[str] | None = None):
    args = parse_args(initial_args)
    setup_logging(logger, args.verbose)

    loop = asyncio.get_event_loop()
    if args.verbose:
        loop.set_debug(True)
    loop.run_until_complete(listener_loop(server=get_server(args)))
    try:
        loop.run_forever()
    finally:
        loop.close()
    logger.info(f"Stopped listening to {args.server}")


if __name__ == "__main__":
    main()
