import argparse
import json
import sys

import requests
from prettytable import PrettyTable

from kamikaze3.common.api import HttpApiClient
from kamikaze3.common.args import OUTPUT_FORMAT_JSON, add_default_args
from kamikaze3.common.env import get_app_name
from kamikaze3.object_storage import objects
from kamikaze3.object_storage.settings import (
    OBJECT_STORAGE_API_ENDPOINT,
    OBJECT_STORAGE_API_WITH_CREDENTIALS_ENDPOINT,
    OBJECT_STORAGE_CANCEL_API_ENDPOINT,
    OBJECT_STORAGE_PRODUCTS_API_ENDPOINT,
)


def fetch_plans():
    try:
        client = HttpApiClient()
        response = client.get(OBJECT_STORAGE_PRODUCTS_API_ENDPOINT)
        return response["results"]
    except requests.RequestException as e:
        print(f"Failed to fetch plans: {e}")
        sys.exit(1)


def select_plan(plans):
    print("Select a plan:")
    for i, plan in enumerate(plans, 1):
        price_eur = plan["price"] / 100
        print(
            f"  {i}. {plan['code']} (Storage: {plan['storage_size_in_gb']}GB, Price: €{price_eur:.2f}/month)"
        )

    while True:
        try:
            choice = int(input("Enter your choice (1-4): "))
            if 1 <= choice <= len(plans):
                return plans[choice - 1]
            else:
                print("Invalid choice. Please select a valid option.")
        except ValueError:
            print("Please enter a valid number.")


def confirm_creation(name, selected_plan):
    print("Object storage creation details:")
    print(f"- Name: {name}")
    print(f"- Plan: {selected_plan['code']}")
    print(f"- Storage Size: {selected_plan['storage_size_in_gb']}GB")
    print(f"- Price: €{selected_plan['price'] / 100:.2f}/month")
    confirm = input("Do you want to proceed? (yes/no): ").strip().lower()
    return confirm == "yes"


def create_storage_request(
    name: str, selected_plan: dict, coupon_code: str | None = None
):
    payload = {
        "name": name,
        "product": selected_plan["code"],
        "apps": [get_app_name()],
    }

    if coupon_code:
        payload["coupon_code"] = coupon_code

    try:
        client = HttpApiClient()
        client.post(OBJECT_STORAGE_API_ENDPOINT, data=payload)
        print(f'Object storage "{name}" created successfully!')
    except requests.RequestException as e:
        print(f"Failed to create object storage: {e.response.json()}")
        sys.exit(1)


def get_plan_by_code(plans, plan_code):
    for plan in plans:
        if plan_code == plan["code"]:
            return plan


def create_object_storage(name=None, plan_code=None, coupon_code=None):
    if not name:
        name = input("Enter a name for the object storage: ").strip()
    plans = fetch_plans()

    selected_plan = None
    if plan_code:
        selected_plan = get_plan_by_code(plans, plan_code)
    if not selected_plan:
        selected_plan = select_plan(plans)

    if confirm_creation(name, selected_plan):
        create_storage_request(name, selected_plan, coupon_code)
    else:
        print("Object storage creation cancelled.")
        sys.exit(1)


def get_current_object_storage_workspace(with_credentials=False):
    url = (
        OBJECT_STORAGE_API_WITH_CREDENTIALS_ENDPOINT
        if with_credentials
        else OBJECT_STORAGE_API_ENDPOINT
    )
    try:
        client = HttpApiClient()
        response = client.get(url)
    except requests.RequestException as e:
        print(f"Failed to fetch object storage workspace: {e}")
        sys.exit(1)

    workspaces = response["results"]

    if not workspaces:
        print(
            "No object storage workspace found. Please create one by running 'hypernode-object-storage create'."
        )
        return None

    workspace = workspaces[0]
    return workspace


def get_object_storage_workspace_info(with_credentials=False):
    workspace = get_current_object_storage_workspace(with_credentials=with_credentials)

    if workspace:
        hidden_credentials = {
            "management_url": "**sensitive**",
            "access_key": "**sensitive**",
            "secret_key": "**sensitive**",
        }

        data = {
            "UUID": workspace["uuid"],
            "Name": workspace["name"],
            "Plan": workspace["product"],
            "Hypernodes": ", ".join(workspace["apps"]),
            "Credentials": (
                workspace["credentials"] if with_credentials else hidden_credentials
            ),
        }

        return data


def print_workspace_info_table(data):
    table = PrettyTable()
    table.field_names = [
        "UUID",
        "Name",
        "Plan",
        "Hypernodes",
        "Management URL",
        "Access Key",
        "Secret Key",
    ]
    table.add_row(
        [
            data["UUID"],
            data["Name"],
            data["Plan"]["code"],
            data["Hypernodes"],
            data["Credentials"]["management_url"],
            data["Credentials"]["access_key"],
            data["Credentials"]["secret_key"],
        ]
    )

    print(table)


def cancel_object_storage():
    workspace = get_current_object_storage_workspace()

    if not workspace:
        sys.exit(1)

    try:
        client = HttpApiClient()
        client.post(OBJECT_STORAGE_CANCEL_API_ENDPOINT.format(workspace["uuid"]))
        print(
            f'Object storage "{workspace["name"]}" has been cancelled. You can '
            f"continue using it until the end of the month, after which it will be deleted."
        )
    except requests.RequestException as e:
        print(
            f"Failed to cancel object storage: {e.response.json() if e.response else e.args[0]}"
        )
        sys.exit(1)


def parse_args(args):
    parser = argparse.ArgumentParser(description="Manage Hypernode Object Storage.")
    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    subparsers.add_parser(
        "cancel",
        help="Cancel current object storage linked to hypernode",
        description="Cancel current object storage linked to hypernode",
    )

    create_parser = subparsers.add_parser(
        "create",
        help="Create new object storage",
        description="Create new object storage",
    )
    create_parser.add_argument(
        "--name", "-n", required=False, help="The name of the object storage"
    )
    create_parser.add_argument(
        "--plan", "-p", required=False, help="The code of the selected plan"
    )
    create_parser.add_argument(
        "--coupon-code", required=False, help="Coupon code to apply"
    )

    info_parser = subparsers.add_parser(
        "info", help="Get object storage info", description="Get object storage info"
    )
    info_parser.add_argument(
        "--with-credentials",
        "-C",
        required=False,
        help="Show sensitive credentials of the workspace",
        action="store_true",
    )
    add_default_args(info_parser)

    objects_parser = subparsers.add_parser(
        "objects",
        help="Manage objects in object storage",
        description="Manage objects in object storage",
    )
    objects.add_arguments_to_parser(objects_parser)
    return parser.parse_known_args(args)


def main(argv=None):
    argv = argv or sys.argv[1:]
    args, optionals = parse_args(argv)

    if args.subcommand == "create":
        create_object_storage(args.name, args.plan, args.coupon_code)
    elif args.subcommand == "info":
        result = get_object_storage_workspace_info(args.with_credentials)
        if not result:
            return

        if args.output_format == OUTPUT_FORMAT_JSON:
            print(json.dumps(result))
        else:
            print_workspace_info_table(result)
    elif args.subcommand == "cancel":
        cancel_object_storage()
    elif args.subcommand == "objects":
        objects.handle(args, optionals)


if __name__ == "__main__":
    main()
