import os
import signal
import subprocess
import sys
from argparse import ArgumentParser, HelpFormatter
from tempfile import NamedTemporaryFile

import psutil

from kamikaze3.common.env import get_hn_object_storage_var
from kamikaze3.object_storage.settings import AWS_DEFAULT_REGION


class CustomFormatter(HelpFormatter):
    def _format_usage(self, usage, actions, groups, prefix):
        return (
            super()._format_usage(usage, actions, groups, prefix).rstrip()
            + " [optional s3 args...]\n\n"
        )


def handle_sigint(signum, frame):
    print("Interrupted")
    sys.exit(0)


signal.signal(signal.SIGINT, handle_sigint)


def add_arguments_to_parser(parser: ArgumentParser):
    subcommand_parser = parser.add_subparsers(dest="objects_subcommand", required=True)
    sync_parser = subcommand_parser.add_parser(
        "sync",
        help="Synchronize files between a local directory and an object storage location",
        formatter_class=CustomFormatter,
    )
    sync_parser.add_argument(
        "source", help="Local directory or S3 URI to sync files from"
    )
    sync_parser.add_argument(
        "destination", help="Local directory or S3 URI to sync files to"
    )

    cp_parser = subcommand_parser.add_parser(
        "cp",
        help="Copy a file or object from one location to another",
        formatter_class=CustomFormatter,
    )
    cp_parser.add_argument("source", help="Path or S3 URI of the source file/object")
    cp_parser.add_argument(
        "destination",
        help="Path or S3 URI of the destination to copy the file/object to",
    )

    ls_parser = subcommand_parser.add_parser(
        "ls",
        help="List objects in an S3 bucket or folder",
        formatter_class=CustomFormatter,
    )
    ls_parser.add_argument(
        "s3_uri",
        help="S3 URi of the bucket or folder to list objects from",
        nargs="?",
        default="",
    )

    mv_parser = subcommand_parser.add_parser(
        "mv",
        help="Move or rename a file or object",
        formatter_class=CustomFormatter,
    )
    mv_parser.add_argument("source", help="Path or S3 URI of the source file/object")
    mv_parser.add_argument(
        "destination",
        help="Path or S3 URI of the destination to move/rename the file/object to",
    )

    rm_parser = subcommand_parser.add_parser(
        "rm",
        help="Delete an object from S3",
        formatter_class=CustomFormatter,
    )
    rm_parser.add_argument("s3_uri", help="S3 URI of the object to delete")

    show_parser = subcommand_parser.add_parser(
        "show",
        help="Display the current status of an ongoing sync process",
    )
    show_parser.add_argument(
        "pid", help="Process ID (PID) of the sync operation to display", type=int
    )


def get_s3_env(config_path: str = None) -> dict[str, str]:
    try:
        env = {
            "AWS_ACCESS_KEY_ID": get_hn_object_storage_var("access_key"),
            "AWS_SECRET_ACCESS_KEY": get_hn_object_storage_var("secret_key"),
            "AWS_ENDPOINT_URL": get_hn_object_storage_var("endpoint_url"),
            "AWS_DEFAULT_REGION": AWS_DEFAULT_REGION,
            "AWS_DEFAULT_OUTPUT": "json",
        }
    except FileNotFoundError:
        env = {}

    if config_path:
        env["AWS_CONFIG_FILE"] = config_path

    return env


def sync(source: str, destination: str, extra_opts: list[str] | None = None) -> int:
    print(f"Syncing objects from {source} to {destination}...")

    cmd = [
        "/usr/bin/aws",
        "s3",
        "sync",
        source,
        destination,
        *extra_opts,
    ]

    with NamedTemporaryFile(mode="w") as aws_config:
        aws_config.write("[s3]\nmax_concurrent_requests = 40\n")
        process = subprocess.Popen(
            cmd,
            env=get_s3_env(aws_config.name),
            text=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            start_new_session=True,
        )

    process_id = process.pid
    try:
        process.wait(timeout=1)
    except subprocess.TimeoutExpired:
        # process is still running, that is good :D
        pass
    else:
        if (rc := process.returncode) != 0:
            stdout, _ = process.communicate()
            stdout = stdout.replace("aws s3", "hypernode-object-storage objects")
            print(f"Failed to start sync process.")
            print(stdout, end="")
            return rc

    print(f"Sync process started with PID {process_id} in the background.")

    return os.EX_OK


def show(pid: int):
    try:
        process = psutil.Process(pid)
        rc = process.wait(timeout=0)
        print(f"Process {pid} exited with code: {rc}")
    except psutil.NoSuchProcess:
        print(f"Process {pid} does not exist anymore")
    except psutil.TimeoutExpired:
        # we routed stderr to stdout previously so this should print both
        cmd = [f"/usr/bin/cat", f"/proc/{pid}/fd/1"]
        subprocess.run(cmd)
    return os.EX_OK


def run_aws_s3_command(*args) -> int:
    cmd = ["/usr/bin/aws", "s3", *args]

    process = subprocess.Popen(
        cmd,
        env=get_s3_env(),
        text=True,
        stdout=None,
        stderr=subprocess.PIPE,
    )

    process.wait()

    if stderr := process.stderr.read():
        stderr = stderr.replace("aws s3", "hypernode-object-storage objects")
        print(stderr, end="")

    return os.EX_OK


def handle(args, optionals):
    match args.objects_subcommand:
        case "sync":
            return sync(args.source, args.destination, optionals)
        case "show":
            return show(args.pid)
        case _:
            posargs = (
                [args.s3_uri]
                if hasattr(args, "s3_uri")
                else [args.source, args.destination]
            )
            return run_aws_s3_command(args.objects_subcommand, *posargs, *optionals)


def parse_args(args: list[str] | None = None):
    parser = ArgumentParser(description="Manage objects in object storage.")
    add_arguments_to_parser(parser)
    return parser.parse_known_args(args)


if __name__ == "__main__":
    sys.exit(handle(parse_args()))
