import json
import subprocess
from copy import deepcopy
from glob import glob
from logging import getLogger
from os import chmod, remove
from pathlib import Path
from platform import system
from uuid import uuid4

from kamikaze3.cluster.health import has_parent_with_relation_type
from kamikaze3.common.env import get_hn_app_var
from kamikaze3.privileged.manage_vhosts import program_name, settings

# Statefile options
from kamikaze3.settings import HYPERNODE_PRIV_GROUP

logger = getLogger(program_name)


def get_persistent_statefile() -> Path:
    app_type = get_hn_app_var("app_type", default_val="persistent")
    if (
        has_parent_with_relation_type("loadbalancer")
        and app_type != settings.HORIZONTAL_AUTOSCALE_APP_TYPE
    ):
        return Path(settings.SHARED_STATE_FILE_PATH)
    return Path(settings.STATE_FILE_PATH)


def remove_lingering_site_configs():
    """
    Find all configs in /etc/nginx/sites and remove them if they are not in the new config.
    This ensures the state of the /etc/nginx/sites directory
    :param dict new_config: New configuration which should be present on the server
    :rtype: None
    """
    expected_files = []
    for server_name in get_all_configs():
        config_name = server_name_to_config_name(server_name)
        expected_files.append(settings.HTTP_SITE_CONFIG.format(config_name))
        expected_files.append(settings.HTTPS_SITE_CONFIG.format(config_name))
        expected_files.append(settings.VARNISH_SITE_CONFIG.format(config_name))

    for file in glob(f"{settings.ROOT_SITES_DIR}/*"):
        if file not in expected_files:
            logger.info(f"Removing lingering config {file}")
            remove(file)


def get_all_configs():
    """
    Retrieve all current configs
    :rtype: dict[dict, ..]
    """
    try:
        with open(settings.STATE_FILE_PATH) as fp:
            return json.load(fp)
    except FileNotFoundError:
        app_type = get_hn_app_var("app_type", default_val="persistent")
        if (
            has_parent_with_relation_type("loadbalancer")
            and app_type != settings.HORIZONTAL_AUTOSCALE_APP_TYPE
        ):
            raise
        return {}


def get_current_config(servername):
    """
    Get the active configuration for the given servername
    :param str servername: Servername that we configure a VHost for.
    :rtype: dict
    """
    try:
        with open(settings.STATE_FILE_PATH) as fp:
            config = json.load(fp)
            if servername in config:
                logger.debug(f"Found existing config for {servername}")
                return config[servername]
            else:
                logger.debug(
                    f"No existing config for {servername}. Current config: {config}"
                )
                logger.info(
                    "No existing config for {}, starting with default options".format(
                        servername
                    )
                )
                return deepcopy(settings.DEFAULT_VHOST_OPTIONS)
    except FileNotFoundError:
        app_type = get_hn_app_var("app_type", default_val="persistent")
        if (
            has_parent_with_relation_type("loadbalancer")
            and app_type != settings.HORIZONTAL_AUTOSCALE_APP_TYPE
        ):
            raise
        return deepcopy(settings.DEFAULT_VHOST_OPTIONS)


def validate_object_storage(config):
    valid = True
    for servername in config["servernames"]:
        current_config = get_current_config(servername)
        will_use_default_settings = not current_config.get(
            "object_storage_bucket"
        ) and not config.get("object_storage_bucket")
        object_storage_disabled = not get_hn_app_var(
            "object_storage_enabled", default_val=False
        )
        if will_use_default_settings and object_storage_disabled:
            valid = False
            logger.warning(
                f"Trying to enable object storage for {servername} with default settings, but object storage is not enabled. Enable it and try again."
            )
    return valid


def _expand_config(current_config, new_config):
    """
    Expand the new config with missing necessary values
    :param dict current_config: Old config currently in the statefile
    :param dict new_config: New config that is to be saved in the statefile
    :rtype: dict
    """
    expanded_config = new_config.copy()

    enabling_object_storage = new_config.get("object_storage_enabled")
    new_bucket_name_empty = not new_config.get("object_storage_bucket")
    current_bucket_name_empty = not current_config.get("object_storage_bucket")

    # only check bucket name because it implies the state of the url too
    if enabling_object_storage and new_bucket_name_empty and current_bucket_name_empty:
        expanded_config["object_storage_bucket"] = (
            settings.DEFAULT_OBJECT_STORAGE_BUCKET_NAME
        )
        expanded_config["object_storage_url"] = (
            settings.DEFAULT_OBJECT_STORAGE_GATEWAY_URL
        )

    return expanded_config


def _filter_config(config):
    """
    Remove any variables from the config that don't need to be saved
    in the statefile.
    :param dict config: Config that is to be saved in the statefile.
    :rtype: dict
    """
    # The default_server option is handled in manage_default_server
    filter_keys = [
        "servernames",
        "default_server",
        "all",
        "list",
        "list_format",
        "yes",
        "no",
        "automatic_confirmation",
        "verbose",
        "import_dir",
        "import_statefile",
        "object_storage_defaults",
    ]
    filter_values = [None]

    filtered_config = {}
    for k, v in config.items():
        if k not in filter_keys and v not in filter_values:
            filtered_config[k] = v
    return filtered_config


def is_wildcard(servername):
    """
    Determine if a servername is a wildcard
    :param str servername: the servername
    :rtype: bool
    """
    return servername.startswith("*")


def server_name_to_config_name(servername):
    """
    Convert a servername to a config name.
    :param ServerName servername: Servername to convert
    :rtype: ConfigName
    """
    if is_wildcard(servername):
        return servername.replace("*", "wildcard")
    return servername


def _update_wildcard_options(config):
    """
    Adjust config if the servername has a wildcard.
    :param dict config: Config of the servername
    :rtype: dict
    """
    servername = config["servername"]
    config["is_wildcard"] = is_wildcard(servername)
    config["config_name"] = server_name_to_config_name(servername)
    return config


def _is_legit_to_quit(servername):
    """
    Check if the servername can be safely deleted.
    :param str servername: Servername that is about to be deleted
    :rtype: bool
    """
    config = get_current_config(servername)
    return not config.get("default_server")


def update_config(servername, new_config):
    """
    Find differences between the current config and the config that is
    defined in the given args.
    Return the keys that changed with their new values defined in the args.
    :param str servername: Name that should match the config
    :param dict new_config: Config with new values
    :rtype: dict
    """
    current_config = get_current_config(servername)
    logger.debug(f"Current config: {current_config}")
    new_config = _expand_config(current_config, new_config)
    # New config has the leading values
    current_config.update(_filter_config(new_config))

    # Add servername to config to make it easier for templating
    current_config["servername"] = servername
    current_config = _update_wildcard_options(current_config)

    return current_config


def update_statefile(new_servername_config):
    """
    Write the new config to the statefile. Create a statefile if it didn't exist yet.
    :param dict new_servername_config: Updated config for the servername
    :rtype: None
    """
    servername = new_servername_config["servername"]
    current_config = get_all_configs()

    if new_servername_config.get("delete"):
        if not _is_legit_to_quit(servername):
            logger.info(
                "Cannot delete default_server {}. Please first change the default_server to another VHost".format(
                    servername
                )
            )
            return
        if servername in current_config:
            del current_config[servername]
        else:
            logger.info(
                "Attempted to delete {} from the vhosts statefile while it wasn't defined.".format(
                    servername
                )
            )
    else:
        current_config[servername] = new_servername_config

    safe_write_root_file(get_persistent_statefile(), json.dumps(current_config))


def safe_write_root_file(path, contents):
    """
    Write the given contents to a directory that is safe for the app user,
    and give it the right permissions for root before moving it to the correct
    location owned by root.
    :param str|PathLike path: Path of the file
    :param str contents: Contents of the file
    :rtype: None
    """
    tempfile = settings.TMP_ROOT_FILE.format(str(uuid4())[:8])
    with open(tempfile, "w") as fp:
        fp.write(contents)
    chmod(tempfile, 0o664)
    try:
        subprocess.check_call(
            [
                "/usr/bin/sudo",
                "/usr/sbin/chown" if system() == "Darwin" else "/bin/chown",
                f"root:{HYPERNODE_PRIV_GROUP}",
                tempfile,
            ]
        )
        subprocess.check_call(["/usr/bin/sudo", "/bin/mv", tempfile, path])
    except subprocess.CalledProcessError as e:
        logger.exception(f"Could not properly write file to {path}")
        raise e
