import logging
import os
import subprocess
import time

from kamikaze3.settings import SUPERVISOR_PROCESS_TEMPLATE
from kamikaze3.supervisorctl.manage_configs import (
    ensure_destination_dir_exists,
    write_config,
)
from kamikaze3.supervisorctl.system_service import (
    ensure_system_service_running,
    is_system_service_running,
)

SUPERVISOR_ROOT_DIR = "/data/web/supervisor/"
logger = logging.getLogger(__name__)


STATE_STOPPED = "STOPPED"
STATE_STARTING = "STARTED"
STATE_RUNNING = "RUNNING"
STATE_BACKOFF = "BACKOFF"
STATE_STOPPING = "STOPPING"
STATE_EXITED = "EXITED"
STATE_FATAL = "FATAL"
STATE_UNKNOWN = "UNKWOWN"

ACTIVE_STATES = (
    STATE_STARTING,
    STATE_RUNNING,
    STATE_BACKOFF,
    STATE_STOPPING,
    STATE_EXITED,
    STATE_FATAL,
    STATE_UNKNOWN,
)


class ProcessStatus:
    def __init__(self, name: str, state: str, message: str) -> None:
        self.name = name
        self.state = state
        self.message = message

    def is_fatal(self) -> bool:
        return self.state == STATE_FATAL

    def is_active(self) -> bool:
        return self.state in ACTIVE_STATES

    def __repr__(self) -> str:
        return f"SupervisorProcess {self.name} ({self.state})"


def ensure_supervisor_running():
    """
    Ensure the supervisor service is enabled and entered running state.
    :return None:
    """
    ensure_system_service_running(
        service_name="supervisor", hypernode_systemctl_setting_name="supervisor_enabled"
    )

    ensure_destination_dir_exists(SUPERVISOR_ROOT_DIR)


def is_supervisor_running():
    """
    Check if supervisor service running
    :return bool supervisor_running: True if supervisor service running, otherwise False
    """
    return is_system_service_running("supervisor")


def manage_supervisor_process_config(command, process_name):
    """
    Ensure supervisor process config exists and return process name
    :param string command: Command to load into template
    :param string process_name: Name of the process
    :return None:
    """
    destination_path = get_destination_path(process_name)
    if not os.path.exists(destination_path):
        logger.info(f"Placing process config for process with name {process_name}")
        write_config(
            {"command": command, "process_name": process_name},
            SUPERVISOR_PROCESS_TEMPLATE,
            destination_path,
        )


def get_destination_path(process_name):
    """
    Get destination path for supervisor process
    :param string process_name: name of the process
    :return string destination_path: destination path of the supervisor process config
    """
    return "".join([SUPERVISOR_ROOT_DIR, process_name, ".conf"])


def ensure_process_running(process_name):
    """
    Ensure process running
    :param string process_name: Process name
    :return bool process_running: Return true if process running, otherwise False
    """
    reread_supervisor()
    add_process(process_name)

    return is_supervisor_process_running(process_name)


def is_supervisor_process_running(process_name):
    """
    Check if process is running. Retry 5 times to get process in running state, otherwise return False.
    :param str process_name: Name of the process
    :return bool: True if process running, otherwise False
    """
    retry_times = 5
    counter = 0
    while counter < retry_times:
        output = exec_supervisor_command("status", process_name, True)
        if STATE_RUNNING in output:
            return True

        time.sleep(0.2)
        counter += 1

    return False


def reread_supervisor():
    """
    Reread supervisor configs
    :return None:
    """
    subprocess.check_call(["/usr/bin/supervisorctl", "reread"])


def update_supervisor_config(process_name=None):
    """
    Update supervisor process
    :param str|None process_name: Optional name of specific process, if no process_name provided, should use 'all'
    :return None:
    """
    if process_name is None:
        process_name = "all"

    exec_supervisor_command("update", process_name)


def add_process(process_name):
    """
    Add process to supervisor
    :param str|None process_name: Name of process to add.
    :return None:
    """
    exec_supervisor_command("add", process_name)


def start_process(process_name=None):
    """
    Start supervisor process
    :param str|None process_name: Optional name of specific process, if no process_name provided, should use 'all'
    :return None:
    """
    if process_name is None:
        process_name = "all"

    exec_supervisor_command("start", process_name)


def restart_process(process_name=None):
    """
    Restart supervisor process
    :param str|None process_name: Optional name of specific process, if no process_name provided, should use 'all'
    :return None:
    """
    if process_name is None:
        process_name = "all"

    exec_supervisor_command("restart", process_name)


def stop_process(process_name=None):
    """
    Stop supervisor process
    :param str|None process_name: Optional name of specific process, if no process_name provided, should use 'all'
    :return None:
    """
    if process_name is None:
        process_name = "all"

    exec_supervisor_command("stop", process_name)


def get_process_statuses(process_name: str = "all") -> list[ProcessStatus]:
    """
    List process(es)
    :param str process_name: Name of the process
    :return List[ProcessStatus]:
    """
    try:
        output = exec_supervisor_command("status", process_name, return_output=True)
    except subprocess.CalledProcessError as e:
        if e.returncode == 3:
            output = e.output
            logger.debug("Supervisorctl returned exit code 3, treating as success.")
        else:
            raise

    result = []
    for line in output.splitlines():
        fields = line.split(maxsplit=2)
        result.append(
            ProcessStatus(fields[0], fields[1], fields[2] if len(fields) > 2 else "")
        )
    return result


def exec_supervisor_command(action, process_name, return_output=False) -> str:
    """
    Exec supervisorctl process for specific process
    :param str action: Action of process (such as start, add etc.)
    :param str process_name: Name of the process
    :param bool return_output: True if should return output of command, default/otherwise False.
    """
    command = ["/usr/bin/supervisorctl", action, process_name]

    if return_output:
        return subprocess.check_output(command, encoding="utf-8")

    subprocess.check_call(command)
    return ""
